from lxml import etree
import zipfile


class ModelDescription(object):

    def __init__(self):
        self.guid = None
        self.fmiVersion = None
        self.modelName = None
        self.description = None
        self.generationTool = None
        self.generationDateAndTime = None
        self.variableNamingConvention = None
        self.numberOfContinuousStates = None
        self.numberOfEventIndicators = None

        self.defaultExperiment = None

        self.coSimulation = None
        self.modelExchange = None

        self.modelVariables = []


class DefaultExperiment(object):

    def __init__(self):
        self.startTime = None
        self.stopTime = None
        self.tolerance = None


class CoSimulation(object):

    def __init__(self):
        self.modelIdentifier = None


class ModelExchange(object):

    def __init__(self):
        self.modelIdentifier = None


class ScalarVariable(object):

    def __init__(self, name, valueReference):
        self.name = name
        self.valueReference = valueReference
        self.description = None
        self.type = None
        self.start = None
        self.causality = None
        self.variability = None

    def __repr__(self):
        return '%s "%s"' % (self.type, self.name)


def read_model_description(filename):

    with zipfile.ZipFile(filename, 'r') as zf:
        xml = zf.open('modelDescription.xml')
        tree = etree.parse(xml)

    root = tree.getroot()

    modelDescription = ModelDescription()

    modelDescription.guid = root.get('guid')
    modelDescription.fmiVersion = root.get('fmiVersion')
    modelDescription.modelName = root.get('modelName')
    modelDescription.description = root.get('description')
    modelDescription.generationTool = root.get('generationTool')
    modelDescription.generationDateAndTime = root.get('generationDateAndTime')
    modelDescription.variableNamingConvention = root.get('variableNamingConvention')

    if root.get('numberOfEventIndicators') is not None:
        modelDescription.numberOfEventIndicators = int(root.get('numberOfEventIndicators'))

    if modelDescription.fmiVersion == '1.0':
        modelDescription.numberOfContinuousStates = int(root.get('numberOfContinuousStates'))
    else:
        modelDescription.numberOfContinuousStates = len(root.findall('ModelStructure/Derivatives/Unknown'))

    defaultExperiment = root.find('DefaultExperiment')

    if defaultExperiment is not None:
        modelDescription.defaultExperiment = DefaultExperiment()
        modelDescription.defaultExperiment.startTime = float(defaultExperiment.get('startTime'))
        modelDescription.defaultExperiment.stopTime = float(defaultExperiment.get('stopTime'))
        #modelDescription.defaultExperiment.tolerance = float(defaultExperiment.get('tolerance'))

    if modelDescription.fmiVersion == "1.0":

        modelIdentifier = root.get('modelIdentifier')

        if root.find('Implementation') is not None:
            modelDescription.coSimulation = CoSimulation()
            modelDescription.coSimulation.modelIdentifier = modelIdentifier
        else:
            modelDescription.modelExchange = ModelExchange()
            modelDescription.modelExchange.modelIdentifier = modelIdentifier

    else:

        me = root.find('ModelExchange')

        if me is not None:
            modelDescription.modelExchange = ModelExchange()
            modelDescription.modelExchange.modelIdentifier = me.get('modelIdentifier')

        cs = root.find('CoSimulation')

        if cs is not None:
            modelDescription.coSimulation = CoSimulation()
            modelDescription.coSimulation.modelIdentifier = cs.get('modelIdentifier')


    modelVariables = root.find('ModelVariables')

    for variable in modelVariables:

        if variable.get("name") is None:
            continue

        sv = ScalarVariable(name=variable.get('name'), valueReference=int(variable.get('valueReference')))
        sv.description = variable.get('description')
        sv.start = variable.get('start')
        sv.causality = variable.get('causality')

        value = next(variable.iterchildren())
        sv.type = value.tag
        start = value.get('start')

        if start is not None:
            if sv.type == 'Real':
                sv.start = float(start)
            elif sv.type == 'Integer':
                sv.start = int(start)
            elif sv.type == 'Boolean':
                sv.start = start == 'true'
            else:
                sv.start = start

        modelDescription.modelVariables.append(sv)


    return modelDescription



import os
import numpy as np
import matplotlib.pyplot as plt
from collections import Iterable
from fmpy.simulate import simulate
from unittest import TestCase
import unittest
import fmpy
from scipy.interpolate import interp1d
import sys

def simulateFMU(path: str, reportDir=None):

    input_filename = None

    for file in os.listdir(path):

        if file.endswith('.fmu'):
            fmu_filename = file

        if file.endswith('_in.csv'):
            input_filename = file

        if file.endswith('_ref.csv'):
            reference_filename = file

        if file.endswith('_cc.bat'):
            batch_filename = file

    print(path, fmu_filename)

    ref = np.genfromtxt(os.path.join(path, reference_filename), delimiter=',', names=True)

    if input_filename is not None:
        input = np.genfromtxt(os.path.join(path, input_filename), delimiter=',', names=True)
    else:
        input = None

    time = ref['time']

    sample_interval = time[1] - time[0]
    stop_time = time[-1]

    step_size = stop_time / 1000

    if batch_filename is not None:

        with open(os.path.join(path, batch_filename), 'r') as file:
            for line in file:
                args = line.split(' ')
                if '-h' in args:
                    i = args.index('-h')
                    step_size = float(args[i+1])
                    break

    res = simulate(filename=os.path.join(path, fmu_filename),
                   stop_time=stop_time,
                   step_size=step_size,
                   output=ref.dtype.names[1:],
                   sample_interval=sample_interval,
                   input=input)

    names = res.dtype.names[1:]

    outliers = 0

    for name in names:
        t_ref = ref['time']
        y_ref = ref[name]

        t_res = res['time']
        y_res = res[name]

        lower, upper = get_limits(y_ref)

        f_u = interp1d(t_ref, upper, fill_value='extrapolate')
        f_l = interp1d(t_ref, lower, fill_value='extrapolate')

        outside = np.logical_or(y_res < f_l(t_res), y_res > f_u(t_res))

        outliers += np.sum(outside)

    # if outliers > 0:
    #     plot_result(res, ref)

    if reportDir is not None:
        header = ','.join(map(lambda s: '"' + s + '"', res.dtype.names))
        np.savetxt(os.path.join(reportDir, 'result.csv'), res, delimiter=',', header=header, comments='', fmt='%g')
        plot_result(res, ref, filename=os.path.join(reportDir, 'result.png'))

        if outliers < 10:
            passedFile = os.path.join(reportDir, 'passed')
            with open(passedFile, 'a'):
                os.utime(passedFile, None)

        with open(os.path.join(reportDir, 'ReadMe.txt'), 'w') as f:
            f.write("See FMPy documentation for how to run simulate FMUs\n")

    return outliers


def get_limits(y_ref, margin=0.02):

    dy = np.abs(y_ref).max() * margin

    _, S = np.meshgrid(range(5), y_ref, indexing='ij')

    # S[0, :-3] = y_ref[3:]
    S[0, :-2] = y_ref[2:]
    S[1, :-1] = y_ref[1:]
    S[2, :]   = y_ref
    S[3, 1:]  = y_ref[:-1]
    S[4, 2:]  = y_ref[:-2]
    # S[6, 3:]  = y_ref[:-3]

    lower = np.min(S, axis=0) - dy
    upper = np.max(S, axis=0) + dy

    return lower, upper


def plot_result(result, reference=None, filename=None):

    import matplotlib.pylab as pylab

    params = {
            # 'legend.fontsize': 'medium',
            #'figure.figsize': (10, 8),
            'legend.fontsize': 'small',
            'axes.labelsize': 'small',
            #   'axes.titlesize': 'medium',
            'xtick.labelsize': 'small',
            'ytick.labelsize': 'small'
    }

    pylab.rcParams.update(params)

    time = result['time']
    names = result.dtype.names[1:]

    if len(names) > 0:

        fig, axes = plt.subplots(len(names), sharex=True)

        fig.set_facecolor('white')

        if not isinstance(axes, Iterable):
            axes = [axes]

        for ax, name in zip(axes, names):

            ax.grid(b=True, which='both', color='0.8', linestyle='-', zorder=0)

            if reference is not None and name in reference.dtype.names:
                t_ref = reference['time']
                y_ref = reference[name]
                lower, upper = get_limits(y_ref)
                ax.fill_between(t_ref, lower, upper, color=(1, 0.5, 0.5))

            ax.plot(time, result[name], 'b', label='result', zorder=101)

            ax.set_ylabel(name)

            ax.margins(x=0, y=0.1)

        if reference is not None:
            #ax.legend()
            plt.legend(bbox_to_anchor=(0, -0.4, 1., 0.1), loc=8, ncol=2, mode="normal", borderaxespad=0.5)


        fig.set_size_inches(w=12, h=8, forward=True)
        plt.tight_layout()
        fig.subplots_adjust(bottom=0.08)

        if filename is None:
            plt.show()
        else:
            dir, file = os.path.split(filename)
            if not os.path.isdir(dir):
                os.makedirs(dir)
            fig.savefig(filename=filename)


class TestSimulateFMU(TestCase):

    def __init__(self, *args, **kwargs):
        super(TestSimulateFMU, self).__init__(*args, **kwargs)

    @classmethod
    def setUpClass(cls):
        cls.html = open(os.path.join(reportDir, 'report.html'), 'w')
        cls.html.write('<html><body><table>')
        cls.html.write('<tr><th>Model</th><th>Outliers</th></tr>\n')

    @classmethod
    def tearDownClass(cls):
        cls.html.write('</table></body></html>')


def test_generator(baseDir, modelPath):
    def test(self):
        outliers = simulateFMU(os.path.join(baseDir, modelPath), reportDir=os.path.join(reportDir, modelPath))
        self.html.write('<tr><td><a href="' + os.path.join(modelPath, 'result.png') + '">' + modelPath + '</a></td><td>' + str(outliers) + '</td></tr>\n')
        self.assertTrue(outliers < 10, "Too many outliers: " + str(outliers))
    return test


for v in ['TEST_FMUS_DIR', 'REPORT_DIR']:
    if v not in os.environ:
        raise Exception('Missing environment variable ' + v)

testFMUsDirectory = os.environ['TEST_FMUS_DIR']
reportDir = os.environ['REPORT_DIR']

fmiVersions = ['fmi1', 'fmi2']
fmiTypes = ['me', 'cs']
tools = ['Dymola']
# tools = ['DS_FMU_Export_from_Simulink']
# tools = ['FMIToolbox_MATLAB']
# tools = ['MapleSim']
# tools = ['SimulationX']
# tools = ['OpenModelica']
# models = ['fullRobot']
models = ['BooleanNetwork1']
# models = ['ControlledTemperature']
# models = None

for fmiVersion in fmiVersions:

    for fmiType in fmiTypes:

        platformDir = os.path.join(testFMUsDirectory,
                                   'FMI_1.0' if fmiVersion == 'fmi1' else 'FMI_2.0',
                                   'ModelExchange' if fmiType == 'me' else 'CoSimulation',
                                   fmpy.platform)

        for tool in os.listdir(platformDir):

            if tools is not None and not tool in tools:
                continue

            toolDir = os.path.join(platformDir, tool)
            versions = os.listdir(toolDir)
            version = sorted(versions)[-1]
            versionDir = os.path.join(toolDir, version)

            for model in os.listdir(versionDir):

                if models is not None and model not in models:
                    continue

                modelDir = os.path.join(versionDir, model)

                if not os.path.isdir(modelDir):
                    continue

                test = test_generator(testFMUsDirectory, os.path.join('FMI_1.0' if fmiVersion == 'fmi1' else 'FMI_2.0',
                                   'ModelExchange' if fmiType == 'me' else 'CoSimulation',
                                   fmpy.platform, tool, version, model))

                setattr(TestSimulateFMU, 'test_' + fmiVersion + '_' + fmiType + '_' + tool +'_' + model, test)


if __name__ == '__main__':
    unittest.main()



# noinspection PyPep8

import sys
from enum import Enum
from ctypes import *
import _ctypes


# determine the platform
if sys.platform == 'win32':
    platform = 'win'
    sharedLibraryExtension = '.dll'
    calloc = cdll.msvcrt.calloc
    free = cdll.msvcrt.free
    freeLibrary = _ctypes.FreeLibrary
elif sys.platform == 'linux':
    platform = 'linux'
    sharedLibraryExtension = '.so'
    from ctypes.util import find_library
    libc = CDLL(find_library("c"))
    calloc = libc.calloc
    free = libc.free
    freeLibrary = _ctypes.dlclose
else:
    raise Exception("Usupported platfrom: " + sys.platform)

calloc.argtypes = [c_size_t, c_size_t]
calloc.restype = c_void_p

free.argtypes = [c_void_p]

if sys.maxsize > 2**32:
    platform += '64'
else:
    platform += '32'

class FMIType(Enum):
    MODEL_EXCHANGE = 0
    CO_SIMULATION = 1

def fmu_info(filename):
    """ Read the FMI version and supported interfaces from an FMU without extracting it """

    from lxml import etree
    import zipfile

    zf = zipfile.ZipFile(filename, 'r')
    md = zf.open('modelDescription.xml')
    tree = etree.parse(md)
    root = tree.getroot()
    version = root.get('fmiVersion')

    return version, ['Co-Simulation']

from .simulate import simulate



# noinspection PyPep8

import os
from ctypes import *
from itertools import combinations
from lxml import etree
from . import free, freeLibrary, platform, sharedLibraryExtension, calloc, FMIType
from .fmi1 import _FMU
import numpy as np


fmi2Component            = c_void_p
fmi2ComponentEnvironment = c_void_p
fmi2FMUstate             = c_void_p
fmi2ValueReference       = c_uint
fmi2Real                 = c_double
fmi2Integer              = c_int
fmi2Boolean              = c_int
fmi2Char                 = c_char
fmi2String               = c_char_p
fmi2Type                 = c_int
fmi2Byte                 = c_char

fmi2Status = c_int

fmi2CallbackLoggerTYPE         = CFUNCTYPE(None, fmi2ComponentEnvironment, fmi2String, fmi2Status, fmi2String, fmi2String, fmi2String)
fmi2CallbackAllocateMemoryTYPE = CFUNCTYPE(c_void_p, c_size_t, c_size_t)
fmi2CallbackFreeMemoryTYPE     = CFUNCTYPE(None, c_void_p)
fmi2StepFinishedTYPE           = CFUNCTYPE(None, fmi2ComponentEnvironment, fmi2Status)

fmi2ModelExchange = 0
fmi2CoSimulation  = 1

fmi2True  = 1
fmi2False = 0

fmi2StatusKind = c_int
fmi2DoStepStatus       = 0
fmi2PendingStatus      = 1
fmi2LastSuccessfulTime = 2
fmi2Terminated         = 3


def logger(a, b, c, d, e, f):
    print(a, b, c, d, e, f)

def allocateMemory(nobj, size):
    return calloc(nobj, size)

def freeMemory(obj):
    free(obj)

def stepFinished(componentEnvironment, status):
    print(combinations, status)

class fmi2CallbackFunctions(Structure):
    _fields_ = [('logger',               fmi2CallbackLoggerTYPE),
                ('allocateMemory',       fmi2CallbackAllocateMemoryTYPE),
                ('freeMemory',           fmi2CallbackFreeMemoryTYPE),
                ('stepFinished',         fmi2StepFinishedTYPE),
                ('componentEnvironment', fmi2ComponentEnvironment)]

callbacks = fmi2CallbackFunctions()
callbacks.logger               = fmi2CallbackLoggerTYPE(logger)
callbacks.allocateMemory       = fmi2CallbackAllocateMemoryTYPE(allocateMemory)
#callbacks.stepFinished         = fmi2StepFinishedTYPE(stepFinished)
callbacks.freeMemory           = fmi2CallbackFreeMemoryTYPE(freeMemory)
#callbacks.componentEnvironment = None

class fmi2EventInfo(Structure):
    _fields_ = [('newDiscreteStatesNeeded',           fmi2Boolean),
                ('terminateSimulation',               fmi2Boolean),
                ('nominalsOfContinuousStatesChanged', fmi2Boolean),
                ('valuesOfContinuousStatesChanged',   fmi2Boolean),
                ('nextEventTimeDefined',              fmi2Boolean),
                ('nextEventTime',                     fmi2Real)]

variables = {}

class ScalarVariable(object):

    def __init__(self, name, valueReference):
        self.name = name
        self.valueReference = valueReference
        self.description = None
        self.type = None
        self.start = None
        self.causality = None
        self.variability = None


class _FMU2(_FMU):

    def __init__(self, modelDescription, unzipDirectory, instanceName, fmiType):

        super(_FMU2, self).__init__(modelDescription, unzipDirectory, instanceName, fmiType)

        # common FMI 2.0 functions
        self.fmi2Instantiate = getattr(self.dll, 'fmi2Instantiate')
        self.fmi2Instantiate.argtypes = [fmi2String, fmi2Type, fmi2String, fmi2String, POINTER(fmi2CallbackFunctions), fmi2Boolean, fmi2Boolean]
        self.fmi2Instantiate.restype = fmi2ComponentEnvironment

        self.fmi2SetupExperiment          = getattr(self.dll, 'fmi2SetupExperiment')
        self.fmi2SetupExperiment.argtypes = [fmi2Component, fmi2Boolean, fmi2Real, fmi2Real, fmi2Boolean, fmi2Real]
        self.fmi2SetupExperiment.restype  = fmi2Status

        self.fmi2EnterInitializationMode          = getattr(self.dll, 'fmi2EnterInitializationMode')
        self.fmi2EnterInitializationMode.argtypes = [fmi2Component]
        self.fmi2EnterInitializationMode.restype  = fmi2Status

        self.fmi2ExitInitializationMode          = getattr(self.dll, 'fmi2ExitInitializationMode')
        self.fmi2ExitInitializationMode.argtypes = [fmi2Component]
        self.fmi2ExitInitializationMode.restype  = fmi2Status

        self.fmi2GetReal          = getattr(self.dll, 'fmi2GetReal')
        self.fmi2GetReal.argtypes = [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Real)]
        self.fmi2GetReal.restype  = fmi2Status

        self.fmi2GetInteger          = getattr(self.dll, 'fmi2GetInteger')
        self.fmi2GetInteger.argtypes = [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Integer)]
        self.fmi2GetInteger.restype  = fmi2Status

        self.fmi2GetBoolean          = getattr(self.dll, 'fmi2GetBoolean')
        self.fmi2GetBoolean.argtypes = [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Boolean)]
        self.fmi2GetBoolean.restype  = fmi2Status

        self.fmi2SetReal          = getattr(self.dll, 'fmi2SetReal')
        self.fmi2SetReal.argtypes = [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Real)]
        self.fmi2SetReal.restype  = fmi2Status

        self.fmi2SetInteger          = getattr(self.dll, 'fmi2SetInteger')
        self.fmi2SetInteger.argtypes = [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Integer)]
        self.fmi2SetInteger.restype  = fmi2Status

        self.fmi2SetBoolean          = getattr(self.dll, 'fmi2SetBoolean')
        self.fmi2SetBoolean.argtypes = [fmi2Component, POINTER(fmi2ValueReference), c_size_t, POINTER(fmi2Boolean)]
        self.fmi2SetBoolean.restype  = fmi2Status

        self.fmi2Terminate          = getattr(self.dll, 'fmi2Terminate')
        self.fmi2Terminate.argtypes = [fmi2Component]
        self.fmi2Terminate.restype  = fmi2Status

        self.fmi2FreeInstance          = getattr(self.dll, 'fmi2FreeInstance')
        self.fmi2FreeInstance.argtypes = [fmi2Component]
        self.fmi2FreeInstance.restype  = None

    def instantiate(self):

        kind = fmi2ModelExchange if self.fmiType == FMIType.MODEL_EXCHANGE else fmi2CoSimulation

        self.component = self.fmi2Instantiate(self.instanceName.encode('utf-8'),
                                              kind,
                                              self.modelDescription.guid.encode('utf-8'),
                                              self.fmuLocation.encode('utf-8'),
                                              byref(callbacks), fmi2False,
                                              fmi2False)

    def setupExperiment(self, tolerance, startTime, stopTime=None):

        toleranceDefined = tolerance is not None

        if tolerance is None:
            tolerance = 0.0

        stopTimeDefined = stopTime is not None

        if stopTime is None:
            stopTime = 0.0

        status = self.fmi2SetupExperiment(self.component, toleranceDefined, tolerance, startTime, stopTimeDefined,
                                          stopTime)

    def enterInitializationMode(self):
        status = self.fmi2EnterInitializationMode(self.component)
        return status

    def exitInitializationMode(self):
        status = self.fmi2ExitInitializationMode(self.component)
        return status

    def getReal(self, vr):
        value = (fmi2Real * len(vr))()
        status = self.fmi2GetReal(self.component, vr, len(vr), value)
        return list(value)

    def setReal(self, vr, value):
        status = self.fmi2SetReal(self.component, vr, len(vr), value)

    def terminate(self):
        status = self.fmi2Terminate(self.component)

    def freeInstance(self):
        self.fmi2FreeInstance(self.component)

        # unload the shared library
        freeLibrary(self.dll._handle)


class FMU2Model(_FMU2):

    def __init__(self, modelDescription, unzipDirectory, instanceName=None):

        super(FMU2Model, self).__init__(modelDescription, unzipDirectory, instanceName, FMIType.MODEL_EXCHANGE)

        self.eventInfo = fmi2EventInfo()

        nx = modelDescription.numberOfContinuousStates
        nz = modelDescription.numberOfEventIndicators

        self.x  = np.zeros(nx)
        self.dx = np.zeros(nx)
        self.z  = np.zeros(nz)

        self._px  = self.x.ctypes.data_as(POINTER(fmi2Real))
        self._pdx = self.dx.ctypes.data_as(POINTER(fmi2Real))
        self._pz  = self.z.ctypes.data_as(POINTER(fmi2Real))

        self.fmi2NewDiscreteStates = getattr(self.dll, 'fmi2NewDiscreteStates')
        self.fmi2NewDiscreteStates.argtypes = [fmi2Component, fmi2EventInfo]
        self.fmi2NewDiscreteStates.restype = fmi2Status

        self.fmi2EnterContinuousTimeMode = getattr(self.dll, 'fmi2EnterContinuousTimeMode')
        self.fmi2EnterContinuousTimeMode.argtypes = [fmi2Component]
        self.fmi2EnterContinuousTimeMode.restype = fmi2Status

        self.fmi2EnterEventMode = getattr(self.dll, 'fmi2EnterEventMode')
        self.fmi2EnterEventMode.argtypes = [fmi2Component]
        self.fmi2EnterEventMode.restype = fmi2Status

        self.fmi2GetContinuousStates = getattr(self.dll, 'fmi2GetContinuousStates')
        self.fmi2GetContinuousStates.argtypes = [fmi2Component, POINTER(fmi2Real), c_size_t]
        self.fmi2GetContinuousStates.restype = fmi2Status

        self.fmi2SetContinuousStates = getattr(self.dll, 'fmi2SetContinuousStates')
        self.fmi2SetContinuousStates.argtypes = [fmi2Component, POINTER(fmi2Real), c_size_t]
        self.fmi2SetContinuousStates.restype = fmi2Status

        self.fmi2GetDerivatives = getattr(self.dll, 'fmi2GetDerivatives')
        self.fmi2GetDerivatives.argtypes = [fmi2Component, POINTER(fmi2Real), c_size_t]
        self.fmi2GetDerivatives.restype = fmi2Status

        self.fmi2GetEventIndicators = getattr(self.dll, 'fmi2GetEventIndicators')
        self.fmi2GetEventIndicators.argtypes = [fmi2Component, POINTER(fmi2Real), c_size_t]
        self.fmi2GetEventIndicators.restype = fmi2Status

        self.fmi2SetTime = getattr(self.dll, 'fmi2SetTime')
        self.fmi2SetTime.argtypes = [fmi2Component, fmi2Real]
        self.fmi2SetTime.restype = fmi2Status

        self.fmi2CompletedIntegratorStep = getattr(self.dll, 'fmi2CompletedIntegratorStep')
        self.fmi2CompletedIntegratorStep.argtypes = [fmi2Component, fmi2Boolean, POINTER(fmi2Boolean), POINTER(fmi2Boolean)]
        self.fmi2CompletedIntegratorStep.restype = fmi2Status

    def newDiscreteStates(self):
        status = self.fmi2NewDiscreteStates(self.component, self.eventInfo)
        return status

    def enterContinuousTimeMode(self):
        status = self.fmi2EnterContinuousTimeMode(self.component)
        return status

    def enterEventMode(self):
        status = self.fmi2EnterEventMode(self.component)
        return status

    def getContinuousStates(self):
        status = self.fmi2GetContinuousStates(self.component, self._px, self.x.size)

    def setContinuousStates(self):
        status = self.fmi2SetContinuousStates(self.component, self._px, self.x.size)

    def getDerivatives(self):
        status = self.fmi2GetDerivatives(self.component, self._pdx, self.dx.size)

    def getEventIndicators(self):
        status = self.fmi2GetEventIndicators(self.component, self._pz, self.z.size)

    def setTime(self, time):
        status = self.fmi2SetTime(self.component, time)

    def completedIntegratorStep(self, noSetFMUStatePriorToCurrentPoint=fmi2True):
        enterEventMode = fmi2Boolean()
        terminateSimulation = fmi2Boolean()
        status = self.fmi2CompletedIntegratorStep(self.component, noSetFMUStatePriorToCurrentPoint, byref(enterEventMode), byref(terminateSimulation))
        return enterEventMode, terminateSimulation


class FMU2Slave(_FMU2):

    def __init__(self, modelDescription, unzipDirectory, instanceName=None):

        super(FMU2Slave, self).__init__(modelDescription, unzipDirectory, instanceName, FMIType.CO_SIMULATION)

        self.fmi2DoStep          = getattr(self.dll, 'fmi2DoStep')
        self.fmi2DoStep.argtypes = [fmi2Component, fmi2Real, fmi2Real, fmi2Boolean]
        self.fmi2DoStep.restype  = fmi2Status

        self.fmi2GetBooleanStatus          = getattr(self.dll, 'fmi2GetBooleanStatus')
        self.fmi2GetBooleanStatus.argtypes = [fmi2Component, fmi2StatusKind, POINTER(fmi2Boolean)]
        self.fmi2GetBooleanStatus.restype  = fmi2Status

    def doStep(self, currentCommunicationPoint, communicationStepSize, noSetFMUStatePriorToCurrentPoint=fmi2True):
        status = self.fmi2DoStep(self.component, currentCommunicationPoint, communicationStepSize, noSetFMUStatePriorToCurrentPoint)
        return status

    def getBooleanStatus(self, kind):
        value = fmi2Boolean(fmi2False)
        status = self.fmi2GetBooleanStatus(self.component, kind, byref(value))
        return value



from distutils.core import setup

setup(name = "fmpy",
      version = "0.0.2",
      description = "Simulate Functional Mockup Units (FMUs) in Python",
      author = "Torsten Sommer",
      author_email = "torsten.sommer@3ds.com",
      url = "https://github.com/CATIA-Systems/FMPy",
      license = "Standard 3-clause BSD",
      packages = ['fmpy',],
      #long_description = """Simulate Functional Mockup Units (FMUs) in Python.""" ,
)



# noinspection PyPep8

import shutil
import zipfile
from tempfile import mkdtemp
from fmpy.modelDescription import read_model_description
from fmpy.fmi1 import *
from fmpy.fmi2 import *
import numpy as np
from collections import Iterable
import sys
from scipy import interpolate


class Recorder(object):

    def __init__(self, fmu, variableNames=None, interval=None):

        self.fmu = fmu
        self.interval = interval

        md = fmu.modelDescription

        is_fmi1 =  md.fmiVersion == '1.0'

        if is_fmi1:
            self._getReal = fmu.fmi1GetReal
            self._getInteger = fmu.fmi1GetInteger
            self._getBoolean = fmu.fmi1GetBoolean
            self._bool_type = fmi1Boolean
        else:
            self._getReal = fmu.fmi2GetReal
            self._getInteger = fmu.fmi2GetInteger
            self._getBoolean = fmu.fmi2GetBoolean
            self._bool_type = fmi2Boolean

        self.cols = [('time', np.float64)]
        self.rows = []

        self.values = {'Real': [], 'Integer': [], 'Boolean': [], 'String': []}

        for sv in md.modelVariables:
            if (variableNames is None and sv.causality == 'output') or (variableNames is not None and sv.name in variableNames):
                self.values['Integer' if sv.type == 'Enumeration' else sv.type].append((sv.name, sv.valueReference))

        if len(self.values['Real']) > 0:
            real_names, real_vrs = zip(*self.values['Real'])
            self.real_vrs = (c_uint32 * len(real_vrs))(*real_vrs)
            self.real_values = (c_double * len(real_vrs))()
            self.cols += zip(real_names, [np.float64] * len(real_names))
        else:
            self.real_vrs = []

        if len(self.values['Integer']) > 0:
            integer_names, integer_vrs = zip(*self.values['Integer'])
            self.integer_vrs = (c_uint32 * len(integer_vrs))(*integer_vrs)
            self.integer_values = (c_int32 * len(integer_vrs))()
            self.cols += zip(integer_names, [np.int32] * len(integer_names))
        else:
            self.integer_vrs = []

        if len(self.values['Boolean']) > 0:
            boolean_names, boolean_vrs = zip(*self.values['Boolean'])
            self. boolean_vrs = (c_uint32 * len(boolean_vrs))(*boolean_vrs)
            self.boolean_values = ((c_int8 if is_fmi1 else c_int32) * len(boolean_vrs))()
            self.cols += zip(boolean_names, ([np.int8] if is_fmi1 else [np.int32]) * len(boolean_names))
        else:
            self.boolean_vrs = []

    def sample(self, time: float, force: bool = False):

        if not force and self.interval is not None and len(self.rows) > 0:
            last = self.rows[-1][0]
            if time - last < self.interval:
                return

        row = [time]

        if len(self.real_vrs) > 0:
            status = self._getReal(self.fmu.component, self.real_vrs, len(self.real_vrs), self.real_values)
            row += list(self.real_values)

        if len(self.integer_vrs) > 0:
            status = self._getInteger(self.fmu.component, self.integer_vrs, len(self.integer_vrs), self.integer_values)
            row += list(self.integer_values)

        if len(self.boolean_vrs) > 0:
            status = self._getBoolean(self.fmu.component, self.boolean_vrs, len(self.boolean_vrs),
                                      cast(self.boolean_values, POINTER(self._bool_type)))
            row += list(self.boolean_values)

        self.rows.append(tuple(row))

    def result(self):
        return np.array(self.rows, dtype=np.dtype(self.cols))


class Input(object):

    def __init__(self, fmu, signals):

        self.fmu = fmu
        self.signals = signals

        if signals is None:
            return

        is_fmi1 = fmu.modelDescription.fmiVersion == '1.0'

        # get the setters
        if is_fmi1:
            self._setReal = fmu.fmi1SetReal
            self._setInteger = fmu.fmi1SetInteger
            self._setBoolean = fmu.fmi1SetBoolean
            self._bool_type = fmi1Boolean
        else:
            self._setReal = fmu.fmi2SetReal
            self._setInteger = fmu.fmi2SetInteger
            self._setBoolean = fmu.fmi2SetBoolean
            self._bool_type = fmi2Boolean

        self.values = {'Real': [], 'Integer': [], 'Boolean': [], 'String': []}

        for sv in fmu.modelDescription.modelVariables:

            if sv.name is None:
                continue

            if sv.causality == 'input':
                if not sv.name in self.signals.dtype.names:
                    print("Warning: missing input for " + sv.name)
                    continue
                f = interpolate.interp1d(self.signals['time'], self.signals[sv.name], kind='linear' if sv.type == 'Real' else 'zero')
                self.values['Integer' if sv.type == 'Enumeration' else sv.type].append((f, sv.valueReference))

        if len(self.values['Real']) > 0:
            self.real_fun, real_vrs = zip(*self.values['Real'])
            self.real_vrs = (c_uint32 * len(real_vrs))(*real_vrs)
            self.real_values = (c_double * len(real_vrs))()
        else:
            self.real_vrs = []

        if len(self.values['Integer']) > 0:
            self.integer_fun, integer_vrs = zip(*self.values['Integer'])
            self.integer_vrs = (c_uint32 * len(integer_vrs))(*integer_vrs)
            self.integer_values = (c_int32 * len(integer_vrs))()
        else:
            self.integer_vrs = []

        if len(self.values['Boolean']) > 0:
            self.boolean_fun, boolean_vrs = zip(*self.values['Boolean'])
            self.boolean_vrs = (c_uint32 * len(boolean_vrs))(*boolean_vrs)
            self.boolean_values = ((c_int8 if is_fmi1 else c_int32) * len(boolean_vrs))()
        else:
            self.boolean_vrs = []

    def apply(self, time):

        if self.signals is None:
            return

        if len(self.real_vrs) > 0:
            for i, f in enumerate(self.real_fun):
                self.real_values[i] = f(time)
            status = self._setReal(self.fmu.component, self.real_vrs, len(self.real_vrs), self.real_values)

        if len(self.integer_vrs) > 0:
            #status = self.fmu.fmi1GetInteger(self.fmu.component, self.integer_vrs, len(self.integer_vrs), self.integer_values)
            #print(time, self.integer_values[0], status)

            for i, f in enumerate(self.integer_fun):
                self.integer_values[i] = f(time)
            status = self._setInteger(self.fmu.component, self.integer_vrs, len(self.integer_vrs), self.integer_values)
            #print(time, f(time), self.integer_values[0], status)

        if len(self.boolean_vrs) > 0:
            for i, f in enumerate(self.boolean_fun):
                self.boolean_values[i] = f(time)
            status = self._setBoolean(self.fmu.component, self.boolean_vrs, len(self.boolean_vrs),
                                      cast(self.boolean_values, POINTER(self._bool_type)))
            # print(time, f(time), self.boolean_values[0], status)


def simulate(filename, start_time=None, stop_time=None, step_size=None, sample_interval=None, fmiType=None, start_values={}, input=None, output=None):

    modelDescription = read_model_description(filename)

    if fmiType is None:
        # determine the FMI type automatically
        fmiType = FMIType.CO_SIMULATION if modelDescription.coSimulation is not None else FMIType.MODEL_EXCHANGE

    defaultExperiment = modelDescription.defaultExperiment

    if start_time is None:
        if defaultExperiment is not None:
            start_time = defaultExperiment.startTime
        else:
            start_time = 0.0

    if stop_time is None:
        if defaultExperiment is not None:
            stop_time = defaultExperiment.stopTime
        else:
            stop_time = 1.0

    if step_size is None:
        total_time = stop_time - start_time
        step_size = 10 ** (np.round(np.log10(0.09)) - 3)

    unzipdir = mkdtemp()

    # expand the 8.3 paths on windows
    if sys.platform == 'win32':
        import win32file
        unzipdir = win32file.GetLongPathName(unzipdir)

    with zipfile.ZipFile(filename, 'r') as fmufile:
        fmufile.extractall(unzipdir)

    if modelDescription.fmiVersion == '1.0':
        simfun = simulateME1 if fmiType is FMIType.MODEL_EXCHANGE else simulateCS1
    else:
        simfun = simulateME2 if fmiType is FMIType.MODEL_EXCHANGE else simulateCS2

    if sample_interval is None:
        sample_interval = (stop_time - start_time) / 500

    # simulate the FMU
    result = simfun(modelDescription, unzipdir, start_time, stop_time, step_size, input, output, sample_interval)

    # clean up
    shutil.rmtree(unzipdir)

    return result


def simulateME1(modelDescription, unzipdir, start_time, stop_time, step_size, input_signals, output, output_interval):

    fmu = FMU1Model(modelDescription=modelDescription, unzipDirectory=unzipdir)
    fmu.instantiate()

    time = start_time

    fmu.setTime(time)

    input = Input(fmu, input_signals)

    input.apply(time)

    fmu.initialize()

    recorder = Recorder(fmu=fmu, variableNames=output, interval=output_interval)

    prez  = np.zeros_like(fmu.z)

    while time < stop_time:

        input.apply(time)

        fmu.getContinuousStates()
        fmu.getDerivatives()

        tPre = time;
        time = min(time + step_size, stop_time);

        timeEvent = fmu.eventInfo.upcomingTimeEvent != fmi1False and fmu.eventInfo.nextEventTime <= time;

        if timeEvent:
            time = fmu.eventInfo.nextEventTime

        dt = time - tPre

        fmu.setTime(time)

        # forward Euler
        fmu.x += dt * fmu.dx

        fmu.setContinuousStates()

        # check for step event, e.g.dynamic state selection
        stepEvent = fmu.completedIntegratorStep()

        # check for state event
        prez[:] = fmu.z
        fmu.getEventIndicators()
        stateEvent = np.any((prez * fmu.z) < 0)

        # handle events
        if timeEvent or stateEvent or stepEvent:
            fmu.eventUpdate()

        recorder.sample(time)

    recorder.sample(time, force=True)

    fmu.terminate()
    fmu.freeInstance()

    return recorder.result()


def simulateCS1(modelDescription, unzipdir, start_time, stop_time, step_size, input_signals, output, output_interval):

    fmu = FMU1Slave(modelDescription=modelDescription, unzipDirectory=unzipdir)
    fmu.instantiate("rectifier1")
    fmu.initialize()

    input = Input(fmu, input_signals)

    recorder = Recorder(fmu=fmu, variableNames=output, interval=output_interval)

    time = start_time

    while time < stop_time:
        recorder.sample(time)
        input.apply(time)
        status = fmu.doStep(currentCommunicationPoint=time, communicationStepSize=step_size)
        time += step_size

    recorder.sample(time, force=True)

    fmu.terminate()
    fmu.freeInstance()

    return recorder.result()


def simulateME2(modelDescription, unzipdir, start_time, stop_time, step_size, input_signals, output, output_interval):

    fmu = FMU2Model(modelDescription=modelDescription, unzipDirectory=unzipdir)
    fmu.instantiate()
    fmu.setupExperiment(tolerance=None, startTime=start_time)
    fmu.enterInitializationMode()
    fmu.exitInitializationMode()

    # event iteration
    fmu.eventInfo.newDiscreteStatesNeeded = fmi2True
    fmu.eventInfo.terminateSimulation = fmi2False

    while fmu.eventInfo.newDiscreteStatesNeeded == fmi2True and fmu.eventInfo.terminateSimulation == fmi2False:
        # update discrete states
        status = fmu.newDiscreteStates()

    fmu.enterContinuousTimeMode()

    input = Input(fmu, input_signals)

    recorder = Recorder(fmu=fmu, variableNames=output, interval=output_interval)

    prez  = np.zeros_like(fmu.z)

    time = start_time

    recorder.sample(time)

    while time < stop_time:

        fmu.getContinuousStates()
        fmu.getDerivatives()

        tPre = time
        time = min(time + step_size, stop_time)

        input.apply(time)

        timeEvent = fmu.eventInfo.nextEventTimeDefined != fmi2False and fmu.eventInfo.nextEventTime <= time

        if timeEvent:
            time = fmu.eventInfo.nextEventTime

        dt = time - tPre

        fmu.setTime(time)

        # forward Euler
        fmu.x += dt * fmu.dx

        fmu.setContinuousStates()

        # check for state event
        prez[:] = fmu.z
        fmu.getEventIndicators()
        stateEvent = np.any((prez * fmu.z) < 0)

        # check for step event
        stepEvent, terminateSimulation = fmu.completedIntegratorStep()

        if timeEvent or stateEvent or stepEvent != fmi2False:

            # handle events
            fmu.enterEventMode()

            fmu.eventInfo.newDiscreteStatesNeeded = fmi2True
            fmu.eventInfo.terminateSimulation = fmi2False

            # update discrete states
            while fmu.eventInfo.newDiscreteStatesNeeded != fmi2False and fmu.eventInfo.terminateSimulation == fmi2False:
                fmu.newDiscreteStates()

            fmu.enterContinuousTimeMode()

        recorder.sample(time)

    recorder.sample(time, force=True)

    fmu.terminate()
    fmu.freeInstance()

    return recorder.result()


def simulateCS2(modelDescription, unzipdir, start_time, stop_time, step_size, input_signals, output, output_interval):

    fmu = FMU2Slave(modelDescription=modelDescription, unzipDirectory=unzipdir)
    fmu.instantiate()
    fmu.setupExperiment(tolerance=None, startTime=start_time)
    fmu.enterInitializationMode()
    fmu.exitInitializationMode()

    input = Input(fmu, input_signals)

    recorder = Recorder(fmu=fmu, variableNames=output, interval=output_interval)

    time = start_time

    while time < stop_time:
        recorder.sample(time)
        input.apply(time)
        fmu.doStep(currentCommunicationPoint=time, communicationStepSize=step_size)
        time += step_size

    recorder.sample(time, force=True)

    fmu.terminate()
    fmu.freeInstance()

    return recorder.result()


if __name__ == '__main__':

    import sys
    import matplotlib.pyplot as plt


    if len(sys.argv) < 2:
        print("Usage:")
        print("> python fmpy.simulate <FMU>")
        exit(-1)

    filename = sys.argv[1]

    result = simulate(filename=filename)

    time = result['time']
    names = result.dtype.names[1:]

    if len(names) > 0:

        fig, axes = plt.subplots(len(names), sharex=True)

        fig.set_facecolor('white')

        if not isinstance(axes, Iterable):
            axes = [axes]

        for ax, name in zip(axes, names):
            ax.plot(time, result[name])
            ax.set_ylabel(name)
            ax.grid(True)
            ax.margins(y=0.1)

        plt.tight_layout()
        plt.show()



# noinspection PyPep8

import os
import pathlib
import numpy as np
from ctypes import *
from itertools import combinations
from . import free, freeLibrary, platform, sharedLibraryExtension, calloc, FMIType

fmi1Component      = c_void_p
fmi1ValueReference = c_uint
fmi1Real           = c_double
fmi1Integer        = c_int
fmi1Boolean        = c_char
fmi1String         = c_char_p

fmi1True  = b'\x01'
fmi1False = b'\x00'

fmi1UndefinedValueReference = -1

fmi1Status = c_int

fmi1CallbackLoggerTYPE         = CFUNCTYPE(None, fmi1Component, fmi1String, fmi1Status, fmi1String, fmi1String)
fmi1CallbackAllocateMemoryTYPE = CFUNCTYPE(c_void_p, c_size_t, c_size_t)
fmi1CallbackFreeMemoryTYPE     = CFUNCTYPE(None, c_void_p)
#fmi1StepFinishedTYPE           = CFUNCTYPE(None, fmi1Component, fmi1Status)
fmi1StepFinishedTYPE           = c_void_p

class fmi1CallbackFunctions(Structure):
    _fields_ = [('logger',         fmi1CallbackLoggerTYPE),
                ('allocateMemory', fmi1CallbackAllocateMemoryTYPE),
                ('freeMemory',     fmi1CallbackFreeMemoryTYPE),
                ('stepFinished',   fmi1StepFinishedTYPE)]

class fmi1EventInfo(Structure):
    _fields_ = [('iterationConverged',          fmi1Boolean),
                ('stateValueReferencesChanged', fmi1Boolean),
                ('stateValuesChanged',          fmi1Boolean),
                ('terminateSimulation',         fmi1Boolean),
                ('upcomingTimeEvent',           fmi1Boolean),
                ('nextEventTime',               fmi1Real)]

def logger(a, b, c, d, e):
    print(a, b, c, d, e)

def allocateMemory(nobj, size):
    return calloc(nobj, size)

def freeMemory(obj):
    free(obj)

def stepFinished(componentEnvironment, status):
    print(combinations, status)

callbacks = fmi1CallbackFunctions()
callbacks.logger               = fmi1CallbackLoggerTYPE(logger)
callbacks.allocateMemory       = fmi1CallbackAllocateMemoryTYPE(allocateMemory)
callbacks.freeMemory           = fmi1CallbackFreeMemoryTYPE(freeMemory)
#callbacks.stepFinished         = fmi1StepFinishedTYPE(stepFinished)
#callbacks.stepFinished = None


class _FMU(object):

    def __init__(self, modelDescription, unzipDirectory, instanceName, fmiType):

        self.modelDescription = modelDescription
        self.unzipDirectory = unzipDirectory
        self.fmiType = fmiType

        if fmiType == FMIType.MODEL_EXCHANGE:
            self.modelIdentifier = modelDescription.modelExchange.modelIdentifier
        else:
            self.modelIdentifier = modelDescription.coSimulation.modelIdentifier

        self.instanceName = instanceName if instanceName is not None else self.modelIdentifier

        self.fmuLocation = pathlib.Path(self.unzipDirectory).as_uri()

        # load the shared library
        self.dll = cdll.LoadLibrary(os.path.join(unzipDirectory, 'binaries', platform, self.modelIdentifier + sharedLibraryExtension))


class _FMU1(_FMU):

    def __init__(self, modelDescription, unzipDirectory, instanceName, fmiType):

        super(_FMU1, self).__init__(modelDescription, unzipDirectory, instanceName, fmiType)

        # common FMI 1.0 functions
        self.fmi1GetReal          = getattr(self.dll, self.modelIdentifier  + '_fmiGetReal')
        self.fmi1GetReal.argtypes = [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Real)]
        self.fmi1GetReal.restype  = fmi1Status

        self.fmi1GetInteger = getattr(self.dll, self.modelIdentifier  + '_fmiGetInteger')
        self.fmi1GetInteger.argtypes = [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Integer)]
        self.fmi1GetInteger.restype = fmi1Status

        self.fmi1GetBoolean = getattr(self.dll, self.modelIdentifier  + '_fmiGetBoolean')
        self.fmi1GetBoolean.argtypes = [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Boolean)]
        self.fmi1GetBoolean.restype = fmi1Status

        self.fmi1SetReal = getattr(self.dll, self.modelIdentifier  + '_fmiSetReal')
        self.fmi1SetReal.argtypes = [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Real)]
        self.fmi1SetReal.restype = fmi1Status

        self.fmi1SetInteger = getattr(self.dll, self.modelIdentifier  + '_fmiSetInteger')
        self.fmi1SetInteger.argtypes = [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Integer)]
        self.fmi1SetInteger.restype = fmi1Status

        self.fmi1SetBoolean = getattr(self.dll, self.modelIdentifier  + '_fmiSetBoolean')
        self.fmi1SetBoolean.argtypes = [fmi1Component, POINTER(fmi1ValueReference), c_size_t, POINTER(fmi1Boolean)]
        self.fmi1SetBoolean.restype = fmi1Status

    def doStep(self, currentCommunicationPoint, communicationStepSize, newStep=fmi1True):
        status = self.fmi1DoStep(self.component, currentCommunicationPoint, communicationStepSize, newStep)
        return status


class FMU1Slave(_FMU1):

    def __init__(self, modelDescription, unzipDirectory, instanceName=None):

        super(FMU1Slave, self).__init__(modelDescription, unzipDirectory, instanceName, FMIType.CO_SIMULATION)

        # FMI 1.0 Co-Simulation functions
        self.fmi1InstantiateSlave = getattr(self.dll, self.modelIdentifier + '_fmiInstantiateSlave')
        self.fmi1InstantiateSlave.argtypes = [fmi1String, fmi1String, fmi1String, fmi1String, fmi1Real, fmi1Boolean, fmi1Boolean, fmi1CallbackFunctions, fmi1Boolean]
        self.fmi1InstantiateSlave.restype = fmi1Component

        self.fmi1InitializeSlave = getattr(self.dll, self.modelIdentifier + '_fmiInitializeSlave')
        self.fmi1InitializeSlave.argtypes = [fmi1Component, fmi1Real, fmi1Boolean, fmi1Real]
        self.fmi1InitializeSlave.restype = fmi1Status

        self.fmi1DoStep = getattr(self.dll, self.modelIdentifier + '_fmiDoStep')
        self.fmi1DoStep.argtypes = [fmi1Component, fmi1Real, fmi1Real, fmi1Boolean]
        self.fmi1DoStep.restype = fmi1Status

        self.fmi1TerminateSlave = getattr(self.dll, self.modelIdentifier + '_fmiTerminateSlave')
        self.fmi1TerminateSlave.argtypes = [fmi1Component]
        self.fmi1TerminateSlave.restype = fmi1Status

        self.fmi1FreeSlaveInstance = getattr(self.dll, self.modelIdentifier + '_fmiFreeSlaveInstance')
        self.fmi1FreeSlaveInstance.argtypes = [fmi1Component]
        self.fmi1FreeSlaveInstance.restype = fmi1Status

    def instantiate(self, mimeType="application/x-fmu-sharedlibrary", timeout=0, visible=fmi1False,
                    interactive=fmi1False, functions=callbacks, loggingOn=fmi1False):

        self.component = self.fmi1InstantiateSlave(self.instanceName.encode('UTF-8'),
                                                   self.modelDescription.guid.encode('UTF-8'),
                                                   self.fmuLocation.encode('UTF-8'),
                                                   mimeType.encode('UTF-8'),
                                                   timeout,
                                                   visible,
                                                   interactive,
                                                   callbacks,
                                                   loggingOn)

    def initialize(self, tStart=0.0, stopTime=None):
        stopTimeDefined = stopTime is not None
        tStop = stopTime if stopTimeDefined else 0
        status = self.fmi1InitializeSlave(self.component, tStart, stopTimeDefined, tStop)
        return status

    def terminate(self):
        status = self.fmi1TerminateSlave(self.component)

    def freeInstance(self):
        self.fmi1FreeSlaveInstance(self.component)
        # unload the shared library
        freeLibrary(self.dll._handle)


class FMU1Model(_FMU1):

    def __init__(self, modelDescription, unzipDirectory, instanceName=None):

        super(FMU1Model, self).__init__(modelDescription, unzipDirectory, instanceName, FMIType.MODEL_EXCHANGE)

        self.eventInfo = fmi1EventInfo()

        nx = modelDescription.numberOfContinuousStates
        nz = modelDescription.numberOfEventIndicators

        self.x = np.zeros(nx)
        self.dx = np.zeros(nx)
        self.z = np.zeros(nz)

        self._px = self.x.ctypes.data_as(POINTER(fmi1Real))
        self._pdx = self.dx.ctypes.data_as(POINTER(fmi1Real))
        self._pz = self.z.ctypes.data_as(POINTER(fmi1Real))

        # FMI 1.0 Model Exchange functions
        self.fmi1InstantiateModel = getattr(self.dll, self.modelIdentifier + '_fmiInstantiateModel')
        self.fmi1InstantiateModel.argtypes = [fmi1String, fmi1String, fmi1CallbackFunctions, fmi1Boolean]
        self.fmi1InstantiateModel.restype = fmi1Component

        self.fmi1SetTime = getattr(self.dll, self.modelIdentifier + '_fmiSetTime')
        self.fmi1SetTime.argtypes = [fmi1Component, fmi1Real]
        self.fmi1SetTime.restype = fmi1Status

        self.fmi1Initialize = getattr(self.dll, self.modelIdentifier + '_fmiInitialize')
        self.fmi1Initialize.argtypes = [fmi1Component, fmi1Boolean, fmi1Real, fmi1EventInfo]
        self.fmi1Initialize.restype = fmi1Status

        self.fmi1GetContinuousStates = getattr(self.dll, self.modelIdentifier + '_fmiGetContinuousStates')
        self.fmi1GetContinuousStates.argtypes = [fmi1Component, POINTER(fmi1Real), c_size_t]
        self.fmi1GetContinuousStates.restype = fmi1Status

        self.fmi1GetDerivatives = getattr(self.dll, self.modelIdentifier + '_fmiGetDerivatives')
        self.fmi1GetDerivatives.argtypes = [fmi1Component, POINTER(fmi1Real), c_size_t]
        self.fmi1GetDerivatives.restype = fmi1Status

        self.fmi1SetContinuousStates = getattr(self.dll, self.modelIdentifier + '_fmiSetContinuousStates')
        self.fmi1SetContinuousStates.argtypes = [fmi1Component, POINTER(fmi1Real), c_size_t]
        self.fmi1SetContinuousStates.restype = fmi1Status

        self.fmi1CompletedIntegratorStep = getattr(self.dll, self.modelIdentifier + '_fmiCompletedIntegratorStep')
        self.fmi1CompletedIntegratorStep.argtypes = [fmi1Component, POINTER(fmi1Boolean)]
        self.fmi1CompletedIntegratorStep.restype = fmi1Status

        self.fmi1GetEventIndicators = getattr(self.dll, self.modelIdentifier + '_fmiGetEventIndicators')
        self.fmi1GetEventIndicators.argtypes = [fmi1Component, POINTER(fmi1Real), c_size_t]
        self.fmi1GetEventIndicators.restype = fmi1Status

        self.fmi1EventUpdate = getattr(self.dll, self.modelIdentifier + '_fmiEventUpdate')
        self.fmi1EventUpdate.argtypes = [fmi1Component, fmi1Boolean, fmi1EventInfo]
        self.fmi1EventUpdate.restype = fmi1Status

        self.fmi1Terminate = getattr(self.dll, self.modelIdentifier + '_fmiTerminate')
        self.fmi1Terminate.argtypes = [fmi1Component]
        self.fmi1Terminate.restype = fmi1Status

        self.fmi1FreeModelInstance = getattr(self.dll, self.modelIdentifier + '_fmiFreeModelInstance')
        self.fmi1FreeModelInstance.argtypes = [fmi1Component]
        self.fmi1FreeModelInstance.restype = fmi1Status

    def instantiate(self, functions=callbacks, loggingOn=fmi1False):
        self.component = self.fmi1InstantiateModel(self.instanceName.encode('UTF-8'),
                                                   self.modelDescription.guid.encode('UTF-8'),
                                                   callbacks,
                                                   loggingOn)

    def setTime(self, time: float):
        status = self.fmi1SetTime(self.component, time)

    def initialize(self, toleranceControlled=fmi1False, relativeTolerance=0.0):
        status = self.fmi1Initialize(self.component, toleranceControlled, relativeTolerance, self.eventInfo)
        return status

    def getContinuousStates(self):
        status = self.fmi1GetContinuousStates(self.component, self._px, self.x.size)

    def setContinuousStates(self):
        status = self.fmi1SetContinuousStates(self.component, self._px, self.x.size)

    def getDerivatives(self):
        status = self.fmi1GetDerivatives(self.component, self._pdx, self.dx.size)

    def completedIntegratorStep(self):
        stepEvent = fmi1Boolean()
        status = self.fmi1CompletedIntegratorStep(self.component, byref(stepEvent))
        return stepEvent != fmi1False

    def getEventIndicators(self):
        status = self.fmi1GetEventIndicators(self.component, self._pz, self.z.size)

    def eventUpdate(self, intermediateResults=fmi1False):
        status = self.fmi1EventUpdate(self.component, intermediateResults, self.eventInfo)

    def terminate(self):
        status = self.fmi1Terminate(self.component)

    def freeInstance(self):
        self.fmi1FreeModelInstance(self.component)
        # unload the shared library
        freeLibrary(self.dll._handle)
